/* YOODA : Yet another Object Oriented Database
   Copyright (C) 1994  Eric Abecassis

This library is free software; you can redistribute it and/or
modify it under the terms of the GNU Library General Public
License as published by the Free Software Foundation; either
version 2 of the License, or (at your option) any later version.

This library is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
Library General Public License for more details.

You should have received a copy of the GNU Library General Public
License along with this library; if not, write to the Free
Software Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA. */


/*****************************************************************************
*
*  MODULE :             Socket Management
*  SCOPE :              Implements RPC++ through Socket
*  FILE :               sockrpc.cc
*  AUTHORS :            E. Abecassis
*  CREATION DATE :      February 1, 1994
*  COPYRIGHTS :        (C) Apic Systemes, 1994
*  HISTORY :
*
******************************************************************************/

#ifdef __GNUG__
#pragma implementation
#endif

/* IMPORTS *******************************************************************/

#define TRACE

#ifndef sockrpc_HXX
#include "sockrpc.hxx"
#endif
#ifndef connectdinfo_HXX
#include "connectdinfo.hxx"
#endif

#include <sys/param.h>
#include <fcntl.h>
#include <sys/socket.h>
#include <poll.h>
#include <stdio.h>
#include <time.h>
#include <unistd.h>
#include <netdb.h>
#include <errno.h>
#include <signal.h>
#include <sys/utsname.h>

#ifndef SOMAXCONN
#define SOMAXCONN 5
#endif

u_int YOOC_PORT ;

proxy(ConnectionD)* ConnectDaemon ;
SvcServer* SvcServer::_svcBase ;

void taskWaitIO(int fd, int events) ;

SvcSocket::SvcSocket(char* svc, int family) : 
    _fd(-1),
    _async(0), 
    _myport(0), _toport(0), _myaddr(0), _toaddr(0),
    _svcname(strdup(svc)), 
    _family(family),
    _sockn(NULL), 
    _error(FALSE),
    _buffIn(NULL), _pcurrIn(0), _ptopIn(NULL),
    _buffOut(NULL), _pcurrOut(NULL), _ptopOut(NULL),
    _brk(NULL)
{
  hostent *     hostinfo;  

  _buffIn = new char[SOCKBUFSIZE] ;
  _buffOut = new char[SOCKBUFSIZE] ;
  _pcurrIn = _buffIn ;
  _ptopIn = _buffIn ;
  _ptopOut = _buffOut + SOCKBUFSIZE ;
  _pcurrOut = _buffOut ;


  struct utsname hostname ;
  uname(&hostname);
  if( (hostinfo = gethostbyname(hostname.nodename)) == 0 ) {
    _myaddr = INADDR_ANY;
  }
  else {
    _myaddr = ntohl(((in_addr *)(hostinfo->h_addr))->s_addr);
  }
} 


// Close the endpoint if open and release the buffers
SvcSocket::~SvcSocket() 
{ 
  if (_fd != -1) close() ;
  if (_svcname != NULL) { free(_svcname) ; }
  if (_buffIn != NULL) { delete _buffIn ; } 
  if (_buffOut != NULL) { delete _buffOut ; }
  if (_sockn != NULL) { free(_sockn) ; }
}


/*= EXTERNAL FUNCTION =======================================================*/
int SvcSocket::read(int fd, char* buffer, int nbytes)
/*=
[DESCRIPTION :]
  (IN)  
  (OUT) 
[DIAGNOSTIC :]
[CALL :]
=*/
{
  if (_async == 0) {
    return ::read(fd, buffer, nbytes) ;
  } else {    
    int readbytes = ::recv(fd, buffer, nbytes, 0) ;
    while ((readbytes == -1) && (errno == EAGAIN)) {
      taskWaitIO(fd, POLLIN) ;
      readbytes = ::read(fd, buffer, nbytes) ;
    }
    return readbytes ;
  }
}


/*= EXTERNAL FUNCTION =======================================================*/
int SvcSocket::write(int fd, char* buffer, int nbytes)
/*=
[DESCRIPTION :]
  (IN)  
  (OUT) 
[DIAGNOSTIC :]
[CALL :]
=*/
{
  int total = 0 ;
  int writebytes ;
  while (total < nbytes) {
    writebytes = ::send(fd, buffer, nbytes-total, 0) ;
    if (writebytes == -1) {
      if (errno != EAGAIN) {
	return -1 ;
      } else {
	taskWaitIO(fd, POLLOUT) ;
	continue ;
      }
    } else {
      total += writebytes ;
      buffer += writebytes ;
    }
    if (total == nbytes) {
      break ;
    } else {
      taskWaitIO(fd, POLLOUT) ;
    }
  }
  return total ;
}

  
int SvcSocket::createSocket() 
{
  int sock = -1 ;

  if ( (sock = ::socket(_family, SOCK_STREAM, 0)) == -1 ) {
    perror("==> SvcSocket : socket creation impossible :") ;
    return -1 ;
  }

  if (_async) {
    int flag = fcntl(sock, F_GETFL) ;
    fcntl(sock, F_SETFL, flag|O_NONBLOCK) ;
  }

  this->initbuf() ;

  int dummy = 1 ;
  if (_family == AF_INET) {
    if (setsockopt(sock, SOL_SOCKET, SO_REUSEADDR, (char*)&dummy,4) < 0) {
      perror("==> SvcSocket : socket setsockopt impossible") ;
      return -1 ;
    }

    dummy = SOCKBUFSIZE/2 ;
    if (setsockopt(sock, SOL_SOCKET, SO_SNDBUF, (char*)&dummy,4) < 0) {
      perror("==> Svcsocket : socket setsockopt impossible") ;
      return -1 ;
    }

    if (setsockopt(sock, SOL_SOCKET, SO_RCVBUF, (char*)&dummy,4) < 0) {
      perror("==> Svcsocket : socket setsockopt impossible") ;
      return -1 ;
    }
  }

/*
  if (setsockopt(sock, SOL_SOCKET, SO_KEEPALIVE, NULL,0) < 0) {
    perror("==> Svcsocket : socket setsockopt impossible") ;
    return -1 ;
  }
  */
  

  return sock ;
}


// Close the Socket endpoint
int SvcSocket::close() 
{
  if (_fd != -1) {
    if (::close(_fd) == -1) {
      perror("==> SvcSocket : socket close impossible") ;
      _fd = -1 ;
      return -1 ;
    }
    _fd = -1 ;
  }
  return 0 ;
}
 

// Initialise  buffers for input-output
void SvcSocket::initbuf() 
{
  _pcurrIn = _buffIn ;
  _ptopIn = _buffIn ;
  _ptopOut = _buffOut + SOCKBUFSIZE ;
  _pcurrOut = _buffOut ; 
}


// Send the content of the buffer 'buf'
// 'len' is the size in bytes of the buffer
int SvcSocket::sendbuf(char* buf, int len) 
{
  memcpy(_pcurrOut, buf, len) ;
  _pcurrOut += len ;
  if (_pcurrOut == _ptopOut) {
    return overflow() ;
  }
  return 0 ;
}


// Receive the content into the buffer 'buf'
// 'len' is the size in bytes of the receive buffer and the amount of bytes
// expected from the endpoint.
int SvcSocket::recvbuf(char* buf, int len) 
{
  if (_pcurrIn == _ptopIn) {
    if (underflow() == -1) return -1 ;
  }
  memcpy(buf, _pcurrIn, len) ;
  _pcurrIn += len ;
  return 0 ;
}


// Manage overflow in the send buffer
int SvcSocket::overflow() 
{
  int size = (int)((u_long)_pcurrOut - (u_long)_buffOut) ;
  if (size == 0) return 0 ;
  if (_fd == -1) {
    printf("==> SvcSocket : send impossible, socket not connected\n") ;
    return -1 ;
  }
  if (this->write(_fd, _buffOut, size) != size) {
    this->lostConnection() ;
    return -1 ;
  } else {
    _pcurrOut = _buffOut ;
  }
  return 0 ;
}


// Manage underflow in the receive buffer
int SvcSocket::underflow() 
{
  _pcurrIn = _buffIn;
  _ptopIn = _pcurrIn ;
  int recsize ;
  if (_fd == -1) {
    printf("==> SvcSocket : recv impossible, socket not connected\n") ;
    return -1 ;
  }
  recsize = this->read(_fd, _pcurrIn, SOCKBUFSIZE) ;
  if (recsize <= 0) {
    this->lostConnection() ;
    return -1 ;
  } else {
    _ptopIn += recsize ;
  }
  return 0 ;
}


// Fill string pointed by name with the name of the local host
char* SvcSocket::host(char* name)  
{  
  struct utsname hostname ;

  uname(&hostname); 
  strcpy(name, hostname.nodename) ;
  return name ;
}


// Fill string pointed by name with the name of the remote host connected
char* SvcSocket::tohost(char* name)  
{  
  if (_fd == -1) {
    name[0] = 0 ;
  }
  if (_family == AF_UNIX) {
    this->host(name) ;
  } else {
    in_addr     addr;
    hostent *   ent;    
    addr.s_addr = _toaddr;
    ent = gethostbyaddr((char *) &addr, sizeof(addr), AF_INET);
    if( ent ) {
      strcpy(name, ent->h_name) ;
    }
    else {
      name[0] = 0;
    }
  }
  return name ;
}


// Public 'send' through the Socket endpoint. Send the content of 'buf'
int SvcSocket::send(char* buf, int len) 
{
  u_int diff ;
  int sendsize ;
  while (len > 0) {
    diff = (u_long)_ptopOut - (u_long)_pcurrOut ;
    sendsize = (int) (diff > len ? len : diff) ; 
    if (sendbuf(buf, sendsize) == -1) return -1 ;
    len -= sendsize ;
    buf += sendsize ;
  }
  return 0 ;
}


// Public 'recv' through the Socket endpoint. Receive data in 'buf'
int SvcSocket::recv(char* buf, int len) 
{
  u_int diff ;
  int recvsize ;
  while (len > 0) {
    diff = (u_long)_ptopIn - (u_long)_pcurrIn ;
    if (diff == 0) { 
      if (underflow() == -1) return -1 ;
      diff = (u_long)_ptopIn - (u_long)_pcurrIn ;
    }
    recvsize = (int) (diff > len ? len : diff) ; 
    if (recvbuf(buf, recvsize) == -1) return -1 ;
    len -= recvsize ;
    buf += recvsize ;
  }
  return 0 ;
}


// Import inet server 
int SvcClient::importInet()
{
  struct sockaddr_in   addr;
  addr.sin_addr.s_addr = htonl(_toaddr);
  addr.sin_port       = (u_short)htons(_toport);
  addr.sin_family     = AF_INET;
  if( ::connect(_fd, (sockaddr *) &addr, sizeof(addr)) == -1 ) {
    perror("==> SvcSocket : socket connection impossible") ;
    return -1 ;
  }
  return 0 ;
}


// Import unix server
int SvcClient::importUnix() 
{
  struct sockaddr_un   addr;
  strcpy(addr.sun_path, this->_sockn) ;
  addr.sun_family     = AF_UNIX;
  if( ::connect(_fd, (sockaddr *) &addr, sizeof(addr)) == -1 ) {
    perror("==> SvcSocket : socket connection impossible") ;
    return -1 ;
  }
  return 0 ;
}


// Redo import for connectionless simulation mode
int SvcClient::reimport() 
{
  if ((_fd = createSocket()) == -1) {
    return -1 ;
  }
  if (_family == AF_INET) {
    return this->importInet() ;
  } else {
    return this->importUnix() ;
  }
}


// Import inet server 
int SvcClient::import(char* host, u_int port) 
{
  hostent *     hostinfo;  

  if (this->isConnected()) this->close() ;
  
  if ((_fd = createSocket()) == -1) {
    return -1 ;
  }
  if( (hostinfo = gethostbyname(host)) == 0 ) {
    perror("==> unknown host") ;
    return -1 ;
  }
  _toaddr = ntohl(((in_addr *)(hostinfo->h_addr))->s_addr);
  _toport = port;  
  return this->importInet() ;
}


// Import Unix socket server if family == AF_UNIX
// Import Inet socket server with filter if family == AF_INET
int SvcClient::import(char* filter) 
{
  if (this->isConnected()) this->close() ;
  
  if ((_fd = createSocket()) == -1) {
    return -1 ;
  }
  if (this->_family == AF_INET) {
    int res ;
    char* host ;
    u_int port ;
    int instance = 1 ;
    if (ConnectDaemon->askServer(_svcname, filter, instance, host, port)==0) {
      res = -1 ;
    } else {
      res = this->import(host, port) ;
    }
    delete host ;    
    return res ;
  } else {
    if (_sockn != NULL) free(_sockn) ;
    _sockn = strdup(filter) ;
    return this->importUnix() ;
  }
}


// Import inet server in connectionless simulated mode
int SvcClient::importCless(char* host, u_int port) 
{
  int res = this->import(host, port) ;
  this->close() ;
  return res ;
}


// Import Unix socket server if family == AF_UNIX
// Import Inet socket server with filter if family == AF_INET
int SvcClient::importCless(char* filter) 
{
  int res = this->import(filter) ;
  this->close() ;
  return res ;
}


// Launch when connection with server is broken
void SvcClient::lostConnection() {

  this->close() ;
  if (_brk != NULL) _brk(this) ;
}


int SvcServer::export(char* filter) 
{
  if (_sockn != NULL) free(_sockn) ;
  _sockn = strdup(filter) ;

  if ((_listenFd = createSocket()) == -1) {
    return -1 ;
  }

  struct sockaddr*   addr;
  int len ;
  if (this->_family == AF_UNIX) {
    len = sizeof(sockaddr_un) ;
    addr = (sockaddr*)new sockaddr_un ;
    strcpy(((sockaddr_un*)addr)->sun_path, _sockn) ;
    ((sockaddr_un*)addr)->sun_family = AF_UNIX;
  } else {
    len = sizeof(sockaddr_in) ;
    addr = (sockaddr*)new sockaddr_in ;
    ((sockaddr_in*)addr)->sin_addr.s_addr = INADDR_ANY;
    ((sockaddr_in*)addr)->sin_port = 0 ;
    ((sockaddr_in*)addr)->sin_family = AF_INET;
  }

  if (bind(_listenFd, addr, len) == -1) {
    this->close();
    perror("==> SvcSocket : socket failed to bind") ;
    return -1 ;
  }

  if (this->_family == AF_INET) {
    if (getsockname(_listenFd, addr, &len) < 0) {
      perror("==> SvcSocket : getsockname failed");
      return -1 ;
    }
    _myport = ntohs(((sockaddr_in*)addr)->sin_port) ;
    ConnectDaemon->beginServer(_svcname,_sockn,_myport, getpid()) ;
    _daemonConnected = TRUE ;
  }

  if (listen(_listenFd, SOMAXCONN) == -1) {
    this->close();
    perror("==> SvcSocket : socket failed to listen") ;
    return -1 ;
  }

  return 0 ;
}


int SvcServer::export(u_int port) 
{
  if ((_listenFd = createSocket()) == -1) {
    return -1 ;
  }

  sockaddr_in   addr;
  addr.sin_addr.s_addr = INADDR_ANY;
  addr.sin_port       = (u_short)htons(port);
  addr.sin_family     = AF_INET;
  
  if (bind(_listenFd, (sockaddr *) &addr, sizeof(addr)) == -1) {
    this->close();
    perror("==> SvcSocket : socket failed to bind") ;
    return -1 ;
  }

  int len = sizeof(sockaddr_in) ;
  if (getsockname(_listenFd, (sockaddr*)&addr, &len) < 0) {
    perror("==> SvcSocket : getsockname failed");
    return -1 ;
  }

  _myport = ntohs(addr.sin_port) ;

  if (listen(_listenFd, SOMAXCONN) == -1) {
    this->close();
    perror("==> SvcSocket : socket failed to listen") ;
    return -1 ;
  }

  return 0 ;
}


int SvcServer::accept() 
{
  if (_family == AF_INET) {
    sockaddr_in   addr;
    int len = sizeof(addr) ;
    if ( (_fd = ::accept(_listenFd, (sockaddr*)&addr, &len)) == -1 ) {
      this->close();
      perror("==> SvcSocket : socket failed to accept") ;
      return -1 ;
    }
    _toaddr = ntohl(addr.sin_addr.s_addr);
    _toport = ntohs(addr.sin_port);
  } else {
    if( (_fd = ::accept(_listenFd, (sockaddr *) 0, 0)) == -1 ) {
      this->close();
      perror("==> SvcSocket : socket failed to accept") ;
      return -1 ; 
    }
  }
  return 0 ;
}


int SvcServer::close() 
{ 
  if (_listenFd != -1) ::close(_listenFd) ;
  if (_daemonConnected) {
    ConnectDaemon->stopServer(_svcname, _sockn, _myport) ;
    _daemonConnected = FALSE ;
  }
  int res = this->SvcSocket::close() ;
  if ( (_sockn !=NULL) && (_family == AF_UNIX) ) {
    unlink(_sockn); free(_sockn) ;
  }
  return res ;
}


void SvcServer::execRpc() 
{
  int numProc ;
  *this >> numProc ;
  services[numProc](*this) ;
}


void SvcServer::lostConnection() 
{
  ::close(this->_fd) ;
  this->_fd = -1 ;
  if (_daemonConnected) {
    ConnectDaemon->freeServer(_svcname, _sockn, _myport) ;
  }
  if (_brk) _brk(this) ;
}


int SvcInit() {
  ConnectDaemon = new proxy(ConnectionD) ;
  char* port = getenv("YOOC_PORT") ;
  if (port == NULL) {
    cerr << "no specified port number for connectd"  << endl ;
    cerr << "use YOOC_PORT environnement variable" << endl ;
    exit(1) ;
  }
  YOOC_PORT = atoi(port) ;   

  struct utsname hostname ;
  uname(&hostname);
  if (ConnectDaemon->importCless(hostname.nodename, YOOC_PORT) == -1) {
    cerr << "import for connection daemon failed" << endl ;
    return -1 ;
  }

  signal(SIGUSR1, SIG_IGN) ;
  ConnectDaemon->connect(getpid()) ;
  return 0 ;
}


